#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "matrixmath.h"
#include "sfm.h"
#include "v3dtex.h"

#include "gctl.h"
#include "sarreality.h"

#include "obj.h"
#include "messages.h"
#include "simcb.h"
#include "simutils.h"
#include "simsurface.h"
#include "simop.h"

#include "sar.h"
#include "config.h"


static double SARSimLandingGearDragCoeff(
	sar_scene_struct *scene, sar_object_struct *obj_ptr,
	sar_obj_landing_gear_struct *lgear_ptr	/* First landing gear. */
);
static double SARSimHoistDragCoeff(
        sar_scene_struct *scene, sar_object_struct *obj_ptr,
        sar_obj_hoist_struct *hoist_ptr
);

double SARSimFindGround(
        sar_scene_struct *scene,
        sar_object_struct **ptr, int total,
        sar_position_struct *pos
);

int SARSimDoMortality(
	sar_core_struct *core_ptr, sar_object_struct *obj_ptr
);

int SARSimRestart(
	sar_core_struct *core_ptr, sar_scene_struct *scene,
        sar_object_struct ***ptr, int *total,
	int obj_num, sar_object_struct *obj_ptr
);
void SARSimDeleteObjectsEffects(
        sar_core_struct *core_ptr, sar_scene_struct *scene,
        sar_object_struct ***ptr, int *total,
        int obj_num
);
void SARSimSetAircraftCrashed(
        sar_scene_struct *scene,
	sar_object_struct ***ptr, int *total,
        sar_object_struct *obj_ptr,
        sar_object_aircraft_struct *obj_aircraft_ptr
);

void SARSimSetSFMValues(
        sar_scene_struct *scene, sar_object_struct *obj_ptr
);
void SARSimGetSFMValues(
        sar_scene_struct *scene, sar_object_struct *obj_ptr
);

int SARSimApplyNaturalForce(
        sar_core_struct *core_ptr, sar_object_struct *obj_ptr
);
int SARSimApplyArtificialForce(
        sar_core_struct *core_ptr, sar_object_struct *obj_ptr
);

void SARSimApplyGCTL(
        sar_core_struct *core_ptr, sar_object_struct *obj_ptr
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))


/*
 *	Returns the drag coeff caused by the state of the first
 *	landing gear on the given object.
 *
 *	Return value is from 0.0 to 1.0 where 0.0 is no additional
 *	drag and 1.0 is full drag.
 */
static double SARSimLandingGearDragCoeff(
	sar_scene_struct *scene, sar_object_struct *obj_ptr,
        sar_obj_landing_gear_struct *lgear_ptr  /* First landing gear. */
)
{
	if(lgear_ptr == NULL)
	    return(0.0);

	/* Is landing gear retractable? If it is not then no
	 * drag should be added.
	 */
	if(lgear_ptr->flags & SAR_LGEAR_FLAG_FIXED)
	    return(0.0);

	/* Return the position of the gear by its animated position,
	 * where 0 is up and (sar_grad_anim_t)-1 is down.
	 */
	return(
	    (double)lgear_ptr->anim_pos /
	    (double)((sar_grad_anim_t)-1) * 0.3
	);
}

/*
 *      Returns the drag coeff caused by the position of the hoist
 *	(how far the rescue basket is from the hoist).
 *
 *      Return value is from 0.0 to 1.0 where 0.0 is no additional
 *      drag and 1.0 is full drag.
 */            
static double SARSimHoistDragCoeff(
        sar_scene_struct *scene, sar_object_struct *obj_ptr,
        sar_obj_hoist_struct *hoist_ptr
)
{
	double rope_cur, rope_max;

	if(hoist_ptr == NULL)
	    return(0.0);

	/* Get rope lengths which imply position of rescue basket.
	 * Remember to subtract the upper z bounds off the lengths.
	 */
	rope_cur = MAX(
	    hoist_ptr->rope_cur - hoist_ptr->contact_z_min, 0.0
	);
	rope_max = MAX(
	    hoist_ptr->rope_max - hoist_ptr->contact_z_min, 0.0
	);

	if(rope_max > 0.0)
	    return(MIN(rope_cur / rope_max, 1.0) * 0.9);
	else
	    return(0.0);
}


/*
 *      Returns the elevation of (not to) the solid ground with respect
 *      to the given position pos. The z coordinate for the given pos
 *      will not be used since this function always returns solid ground.
 *
 *      Only objects of type ground will be checked, becareful not
 *      to check the given pos if it came from a ground object or
 *      else it would return its own elevation.
 *
 *      So this function should only be used to check landable/walkable
 *      ground for objects that need it (ie aircraft and humans).
 *      Also note that pos passed to this function should not come from
 *      an object who's contact bounds specify the crash flag
 *      SAR_CRASH_FLAG_SUPPORT_SURFACE or else they will keep being
 *      supported on themselves.
 */
double SARSimFindGround(
        sar_scene_struct *scene,
        sar_object_struct **ptr, int total,
        sar_position_struct *pos 
)
{ 
        int i;
        sar_object_struct *tar_obj_ptr;
        double new_height, cur_height = 0.0;
        sar_contact_bounds_struct *cb_tar;
        sar_position_struct *pos_src = pos;
 

        if((scene == NULL) ||
           (pos_src == NULL)
        )
            return(cur_height);
  
        /* Got through last to first object. */
        for(i = total - 1; i >= 0; i--)
        {
            tar_obj_ptr = ptr[i];
            if(tar_obj_ptr == NULL)
                continue;
 
            /* Check if target object specifies landable surface */
            cb_tar = tar_obj_ptr->contact_bounds;
            if(((cb_tar == NULL) ? 0 : cb_tar->crash_flags) &
                SAR_CRASH_FLAG_SUPPORT_SURFACE
            )
            {
                new_height = SARSimSupportSurfaceHeight(
                    tar_obj_ptr, cb_tar, pos_src
                );
                if(new_height > cur_height)
                    cur_height = new_height;
            }
 
            /* Get ground object height (function will check
             * if object really is a ground object), on error
             * it will return 0.0 so accepting its value unconditionally
             * is safe.
             *
             * Both ground object cylendrical contact and heightfield
             * contact will be checked.
             */
            new_height = SARSimHFGetGroundHeight(
                tar_obj_ptr,    /* Ground object. */
                pos_src
            );

            /* New height higher? */
            if(new_height > cur_height)
                cur_height = new_height;
        }

        return(cur_height);
}


/*
 *      Checks if the life span of the given object has been exceeded,
 *      if so then that object will be deleted and returns non-zero.
 *      Calling function may need to consider the pointer array base for
 *	the objects changed if non-zero is returned.
 *
 *	Any object who's life_span is 0 or less will be ignored.
 */
int SARSimDoMortality(
        sar_core_struct *core_ptr,
        sar_object_struct *obj_ptr
)
{
        if((core_ptr == NULL) ||
           (obj_ptr == NULL)
        )
            return(0);

	/* Ignore objects who's life_span is 0 or less. */
        if(obj_ptr->life_span <= 0)
            return(0);

	/* Check if object's life_span has exceeded the current
	 * time. Note that the life_span value was calculated from
	 * when it was created plus its life span (cur_millitime +
	 * life_expectancy) so when the current time exceeds the
	 * object's life_span it indicates the object must be deleted.
	 * All units are in milliseconds.
	 */
        if(obj_ptr->life_span <= cur_millitime)
        {
            int i;

            for(i = 0; i < core_ptr->total_objects; i++)
            {
                if(core_ptr->object[i] == obj_ptr)
                    SARObjDelete(
                        core_ptr,
                        &core_ptr->object,
                        &core_ptr->total_objects,
                        i
                    );
            }
            return(1);
        }
        else 
        {
            return(0);
        }
}

/*
 *	Moves the given object to the nearest helipad mark restartable
 *	and refuels/repairs the object.
 *
 *	This function is intended to be called when the player has
 *	crashed and needs to be restarted.
 *
 *	Does not check if restarting is valid, all permission checks must
 *	be done by the calling function before calling this function.
 *
 *	Returns the index to the helipad object the object is restarting
 *	at or -1 on failure.
 */
int SARSimRestart(
	sar_core_struct *core_ptr, sar_scene_struct *scene,
        sar_object_struct ***ptr, int *total,
	int obj_num, sar_object_struct *obj_ptr
)
{
	int i, closest_obj_num = -1;
	double closest_distance = 0.0;
	sar_position_struct *pos;
	sar_object_struct *start_obj_ptr, *closest_obj_ptr = NULL;
	sar_object_helipad_struct *obj_helipad_ptr = NULL;
	sar_object_aircraft_struct *obj_aircraft_ptr;
	sar_obj_hoist_struct *hoist_ptr;


	if((scene == NULL) || (ptr == NULL) || (total == NULL) ||
           (obj_ptr == NULL)
	)
	    return(-1);

	pos = &obj_ptr->pos;

	/* Check if this is the player object on the scene structure. */
	if((obj_num == scene->player_obj_num) ||
	   (obj_ptr == scene->player_obj_ptr)
	)
	{
	    /* Reset player has crashed mark on scene structure. */
	    scene->player_has_crashed = 0;

	    /* Also reset time compression. */
	    time_compression = 1.0;

	    /* Move camera referance back to cockpit. */
	    scene->camera_ref = SAR_CAMERA_REF_COCKPIT;

	}

        /* Repair and refuel the object. */
        SARSimOpRepair(scene, obj_ptr);
        SARSimOpRefuel(scene, obj_ptr);


	/* Itterate through each object, searching for closest
	 * helipad to the given object.
	 */
	for(i = 0; i < (*total); i++)
	{
	    start_obj_ptr = (*ptr)[i];
	    if(start_obj_ptr == NULL)
		continue;

	    if(start_obj_ptr == obj_ptr)
		continue;

	    if(start_obj_ptr->type != SAR_OBJ_TYPE_HELIPAD)
		continue;
	    else
		obj_helipad_ptr = (sar_object_helipad_struct *)start_obj_ptr->data;

	    /* Skip helipad referanced object. */
	    if(obj_helipad_ptr->flags & SAR_HELIPAD_FLAG_REF_OBJECT)
	    {
		if((obj_helipad_ptr->ref_object == obj_num) &&
                   (obj_num > -1)
		)
		    continue;
	    }

	    /* This helipad allows restarting at? */
	    if(obj_helipad_ptr->flags & SAR_HELIPAD_FLAG_RESTART_POINT)
	    {
		double dx, dy, dlen;

		dx = start_obj_ptr->pos.x - pos->x;
		dy = start_obj_ptr->pos.y - pos->y;
		/* Ignore z. */
		dlen = sqrt((dx * dx) + (dy * dy));

		/* No previous closest object recorded? */
		if(closest_obj_ptr == NULL)
		{
		    closest_distance = dlen;
		    closest_obj_ptr = start_obj_ptr;
		    closest_obj_num = i;
		}
		else
		{
		    /* Closer? */
		    if(dlen < closest_distance)
		    {
			closest_distance = dlen;
                        closest_obj_ptr = start_obj_ptr;
			closest_obj_num = i;
		    }
		}
	    }
	}
	/* Got closest helipad? */
	if(closest_obj_ptr != NULL)
	{
	    /* Move object to closest helipad. */
	    memcpy(pos, &closest_obj_ptr->pos, sizeof(sar_position_struct));

            /* Adjust ground_elevation_msl so that it's at the position of
             * the player object, this way the player won't suddenly drop
             * and be destroyed.
             */
            obj_ptr->ground_elevation_msl = pos->z;

	    /* Set direction to match helipad. */
	    memcpy(&obj_ptr->dir, &closest_obj_ptr->dir, sizeof(sar_direction_struct));

	    /* Need to move object up. */
	    switch(obj_ptr->type)
	    {
	      case SAR_OBJ_TYPE_AIRCRAFT:
		obj_aircraft_ptr = (sar_object_aircraft_struct *)obj_ptr->data;
		if(obj_aircraft_ptr != NULL)
		{
		    /* Reset current speed and velocity. */
		    memset(&obj_aircraft_ptr->vel, 0x00, sizeof(sar_position_struct));
		    obj_aircraft_ptr->speed = 0.0;

		    /* Lower landing gears, first mark as not landed
		     * in order to trick it to lower landing gears.
		     */
		    obj_aircraft_ptr->landed = 0;
		    SARSimOpLandingGear(
			scene, ptr, total,
			obj_ptr, 0
		    );

		    /* Reset hoist. */
		    hoist_ptr = SARObjGetHoistPtr(obj_ptr, 0, NULL);
		    if(hoist_ptr != NULL)
		    {
			hoist_ptr->rope_cur = 0.0;
		    }

		    /* Close door. */
		    SARSimOpDoor(
			scene, ptr, total,
			obj_ptr, 0
		    );

		    /* Turn engine on. */
		    SARSimOpEngine(
			scene, ptr, total,
			obj_ptr, SAR_ENGINE_STATE_ON
		    );

		    /* Move aircraft up just a bit to offset for its 
		     * belly and gear heights.
		     */
		    pos->z += (obj_aircraft_ptr->belly_height +
			obj_aircraft_ptr->gear_height);

		    /* Implicitly mark as landed and not on water. */
		    obj_aircraft_ptr->landed = 1;
		    obj_aircraft_ptr->on_water = 0;
		}
		break;

	    }

	    /* Realize new position. */
	    SARSimWarpObject(
		scene, obj_ptr, pos, &obj_ptr->dir
	    );
	}

	/* Destroy all explosion objects. */
	SARSimDeleteObjectsEffects(
	    core_ptr, scene, ptr, total,
	    obj_num
	);

	return(closest_obj_num);
}


/*
 *	Deletes all explosion, smoke, and other `effects objects'
 *	with respect to the given referance object number.
 *
 *	If the given object is -1 then all of the above will be deleted.
 */
void SARSimDeleteObjectsEffects(
	sar_core_struct *core_ptr, sar_scene_struct *scene,
	sar_object_struct ***ptr, int *total,
	int obj_num
)
{
	int i;
	Boolean need_continue;
	sar_object_struct *effects_obj_ptr;
	sar_object_smoke_struct *obj_smoke_ptr;
	sar_object_explosion_struct *obj_explosion_ptr;


	if((core_ptr == NULL) || (scene == NULL) ||
           (ptr == NULL) || (total == NULL)
	)
	    return;

	for(i = 0; i < (*total); i++)
	{
	    effects_obj_ptr = (*ptr)[i];
	    if(effects_obj_ptr == NULL)
		continue;

	    need_continue = False;

	    switch(effects_obj_ptr->type)
	    {
	      case SAR_OBJ_TYPE_SMOKE:
		obj_smoke_ptr = (sar_object_smoke_struct *)effects_obj_ptr->data;
		if(obj_smoke_ptr != NULL)
		{
		    if(obj_num > -1)
			need_continue = True;
		}
		break;

	      case SAR_OBJ_TYPE_FIRE:
		if(obj_num > -1)       
                    need_continue = True;
		break;

	      case SAR_OBJ_TYPE_EXPLOSION:
                obj_explosion_ptr = (sar_object_explosion_struct *)
		    effects_obj_ptr->data;
                if(obj_explosion_ptr != NULL)
		{
		    if(obj_num > -1)
		    {
			if(obj_explosion_ptr->ref_object != obj_num)
			    need_continue = True;
		    }
                }
		break;

/* Add support for other effects objects. */

	      default:
		/* Do not delete all other types of objects. */
		need_continue = True;
		break;
	    }
	    /* Do not delete this object? */
	    if(need_continue)
		continue;

	    SARObjDelete(
		core_ptr, ptr, total, i
	    );
	}

	return;
}

/*
 *	Sets up values on the given aircraft object to be crashed.
 *
 *	Does not create any explosions, play sounds, or updates related
 *	resources, it is up to the calling function to do that about this
 *	crash.
 */
void SARSimSetAircraftCrashed(
	sar_scene_struct *scene,
	sar_object_struct ***ptr, int *total,
        sar_object_struct *obj_ptr,
	sar_object_aircraft_struct *obj_aircraft_ptr
)
{
	int i, occupant_obj_num;
	sar_direction_struct *dir, new_dir;
	sar_obj_hoist_struct *hoist_ptr;
	sar_object_struct *occupant_obj_ptr;
	sar_object_human_struct *obj_human_ptr;


	if((scene == NULL) || (obj_ptr == NULL) ||
	   (obj_aircraft_ptr == NULL) || (ptr == NULL) || (total == NULL)
	)
	    return;

	/* Get direction of object. */
	dir = &obj_ptr->dir;

	/* Get pointer to aircraft's hoist. */
	hoist_ptr = SARObjGetHoistPtr(obj_ptr, 0, NULL);
	if(hoist_ptr != NULL)
	{
	  /* Itterate through occupants in hoist. */
	  for(i = 0; i < hoist_ptr->total_occupants; i++)
	  {
	    occupant_obj_num = hoist_ptr->occupant[i];
	    if(SARObjIsAllocated(*ptr, *total, occupant_obj_num))
		occupant_obj_ptr = (*ptr)[occupant_obj_num];
	    else
		continue;

	    /* Skip our object itself just in case. */
	    if(occupant_obj_ptr == obj_ptr)
		continue;

	    /* Handle by occupant object type. */
	    switch(occupant_obj_ptr->type)
	    {
	      case SAR_OBJ_TYPE_HUMAN:
                obj_human_ptr = (sar_object_human_struct *)
		    occupant_obj_ptr->data;
                if(obj_human_ptr != NULL)
		{
		    /* Mark human as no longer gripped in the rescue
		     * basket.
		     */
		    obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_GRIPPED;

		    /* Mark as lying down. */
		    obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_SIT;
		    obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_SIT_DOWN;
		    obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_SIT_UP;
		    obj_human_ptr->flags |= SAR_HUMAN_FLAG_LYING;

		    /* Mark human as not aleart and not aware. */
		    obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_ALERT;
		    obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_AWARE;

		    obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_RUN_TOWARDS;
		    obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_RUN_AWAY;
		    obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_PUSHING;
		    obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_DIVER_CATCHER;

		    /* Not in water (need to work on this). */
		    obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_IN_WATER;

		    obj_human_ptr->flags &= ~SAR_HUMAN_FLAG_ON_STREATCHER;
		}
		break;

/* Add support for other types of objects here. */

	    }

	    /* Set attitude to make it `upright' or lie flat. */
	    occupant_obj_ptr->dir.pitch = (0.0 * PI);
	    occupant_obj_ptr->dir.bank = (0.0 * PI);

	    /* Mark this occupant as no longer in the list. */
	    hoist_ptr->occupant[i] = -1;
	  }

	  /* Clear hoist list. */
	  free(hoist_ptr->occupant);
	  hoist_ptr->occupant = NULL;
	  hoist_ptr->total_occupants = 0;

	  /* Retract rope. */
	  hoist_ptr->rope_cur = 0.0;
	}

	/* Begin setting values on aircraft. */

	/* Mark as not flyable. */
	obj_aircraft_ptr->air_worthy_state =
	    SAR_AIR_WORTHY_NOT_FLYABLE;

	/* Turn engine off. */
	obj_aircraft_ptr->engine_state =
	    SAR_ENGINE_STATE_OFF;

	/* Set new direction. */
	new_dir.heading = dir->heading;
	new_dir.pitch = (0.0 * PI);
	new_dir.bank = (1.7 * PI);   /* Tilt it. */

	SARSimWarpObject(scene, obj_ptr, NULL, &new_dir);

	/* Turn off lights. */
	SARSimOpLights(
	    scene, ptr, total, obj_ptr, False
	);
	/* Turn off spot lights. */
	SARSimOpAttenuate(
	    scene, ptr, total, obj_ptr, False
	);
	/* Turn off strobes. */
	SARSimOpStrobes(
	    scene, ptr, total, obj_ptr, False
	);

	return;
}


/*
 *      Sets the flyable object's values to the SFM's values.
 */
void SARSimSetSFMValues(
        sar_scene_struct *scene,
        sar_object_struct *obj_ptr
)
{
	int i;
        SFMModelStruct *fdm = NULL;
        sar_object_aircraft_struct *obj_aircraft_ptr;
        int air_worthy_state;
        sar_obj_landing_gear_struct *lgear_ptr = NULL;
	sar_obj_hoist_struct *hoist_ptr;
        sar_contact_bounds_struct *cb;


        if((scene == NULL) ||
           (obj_ptr == NULL)
        )
            return;

	/* Get pointer to contact bounds, checked if NULL when referanced
	 * later.
	 */
        cb = obj_ptr->contact_bounds;

        switch(obj_ptr->type)
        {
          case SAR_OBJ_TYPE_AIRCRAFT:
            obj_aircraft_ptr = (sar_object_aircraft_struct *)obj_ptr->data;
            if(obj_aircraft_ptr == NULL)
                break;

            fdm = obj_aircraft_ptr->fdm;
            if(fdm == NULL)
                break;
         
            air_worthy_state = obj_aircraft_ptr->air_worthy_state;
	    hoist_ptr = SARObjGetHoistPtr(obj_ptr, 0, NULL);

            /* Get pointer to first landing gear. */
            if(obj_aircraft_ptr->total_landing_gears > 0)
                lgear_ptr = obj_aircraft_ptr->landing_gear[0];

            /* No landing gears if not flyable. */
            if(air_worthy_state == SAR_AIR_WORTHY_NOT_FLYABLE)
                lgear_ptr = NULL;  


            /* Begin setting SFM values from object. */
#define TAR_PTR fdm
#define SRC_PTR obj_aircraft_ptr

            TAR_PTR->flags = (SFMFlagFlightModelType |
                SFMFlagPosition | SFMFlagDirection |
                SFMFlagVelocityVector | SFMFlagSpeed |
                SFMFlagSpeedStall | SFMFlagDragMin |
                SFMFlagSpeedMax | SFMFlagAccelResponsiveness |
                SFMFlagGroundElevation | SFMFlagServiceCeiling |
                SFMFlagBellyHeight | SFMFlagGearState |
                SFMFlagGearType | SFMFlagGearHeight |
                SFMFlagGearBrakesState | SFMFlagGearTurnVelocityOptimul |
                SFMFlagGearTurnVelocityMax | SFMFlagGearTurnRate |
                SFMFlagLandedState | SFMFlagGroundContactType |
                SFMFlagHeadingControlCoeff | SFMFlagBankControlCoeff |
                SFMFlagPitchControlCoeff | SFMFlagThrottleCoeff |
                SFMFlagAfterBurnerState | SFMFlagAfterBurnerPowerCoeff |
                SFMFlagEnginePower | SFMFlagTotalMass |
                SFMFlagAttitudeChangeRate | SFMFlagAttitudeLevelingRate |
                SFMFlagAirBrakesState | SFMFlagAirBrakesRate |
                SFMFlagCanCrashIntoOther | SFMFlagCanCauseCrash |
                SFMFlagCrashContactShape | SFMFlagCrashableSizeRadius |
                SFMFlagCrashableSizeZMin | SFMFlagCrashableSizeZMax |
                SFMFlagTouchDownCrashResistance | SFMFlagCollisionCrashResistance
            );
                
            /* Update flight model type only if SFM not in slew mode. */
	    switch(SRC_PTR->flight_model_type)
	    {
              case SAR_FLIGHT_MODEL_SLEW:
		TAR_PTR->type = SFMFlightModelSlew;
		break;
	      case SAR_FLIGHT_MODEL_AIRPLANE:
		TAR_PTR->type = SFMFlightModelAirplane;
		break;
	      default:
		TAR_PTR->type = SFMFlightModelHelicopter;
                break;
	    }
            TAR_PTR->speed_stall = SRC_PTR->speed_stall;
            TAR_PTR->drag_min = SRC_PTR->min_drag;
	    /* Calculate maximum speed, take drag into account. */
            TAR_PTR->speed_max =
		SRC_PTR->speed_max *
		(1.0 - SARSimHoistDragCoeff(scene, obj_ptr, hoist_ptr)) *
		(1.0 - SARSimLandingGearDragCoeff(scene, obj_ptr, lgear_ptr));
	    TAR_PTR->speed_max_expected = SRC_PTR->speed_max_expected;
	    /* If not flyable, then reduce speed bounds. */
	    if(air_worthy_state == SAR_AIR_WORTHY_NOT_FLYABLE)
	    {
		TAR_PTR->drag_min = TAR_PTR->drag_min * 5.0;
                TAR_PTR->speed_max = TAR_PTR->speed_max / 15.0;
	    }
	    /* Set acceleration responsiveness by flight model type. */
            switch(TAR_PTR->type)
            {
              case SFMFlightModelAirplane:
                TAR_PTR->accel_responsiveness.x =   
                    SRC_PTR->airplane_accel_responsiveness.x; 
                TAR_PTR->accel_responsiveness.y =
                    SRC_PTR->airplane_accel_responsiveness.y;
                TAR_PTR->accel_responsiveness.z =
                    SRC_PTR->airplane_accel_responsiveness.z;
                break;
              default:  /* Assume SFMFlightModelHelicopter. */ 
                TAR_PTR->accel_responsiveness.x =
                    SRC_PTR->accel_responsiveness.x;
                TAR_PTR->accel_responsiveness.y =
                    SRC_PTR->accel_responsiveness.y;   
                TAR_PTR->accel_responsiveness.z =
                    SRC_PTR->accel_responsiveness.z;
                break;
            }
            TAR_PTR->service_ceiling = SRC_PTR->service_ceiling;
            TAR_PTR->belly_height = SRC_PTR->belly_height;
            TAR_PTR->ground_elevation_msl = obj_ptr->ground_elevation_msl;
            TAR_PTR->gear_state = ((lgear_ptr == NULL) ?
                False : !(lgear_ptr->flags & SAR_LGEAR_FLAG_STATE)
            );
            if(lgear_ptr == NULL)
            {
                /* Assume wheels. */
                TAR_PTR->gear_type = SFMGearTypeWheels;
            }
            else
            {
                if(lgear_ptr->flags & SAR_LGEAR_FLAG_SKI)
                    TAR_PTR->gear_type = SFMGearTypeSkis;
                else
                    TAR_PTR->gear_type = SFMGearTypeWheels;
            }
            TAR_PTR->gear_height = SRC_PTR->gear_height;
            TAR_PTR->gear_brakes_state = ((SRC_PTR->wheel_brakes_state > 0) ?
                True : False
            );
            TAR_PTR->gear_turn_velocity_optimul = SRC_PTR->gturn_vel_opt;
            TAR_PTR->gear_turn_velocity_max = SRC_PTR->gturn_vel_max;
/*            TAR_PTR->gear_turn_rate = SRC_PTR->; */
                
/*            TAR_PTR->ground_contact_type = SRC_PTR->; */
            TAR_PTR->heading_control_coeff = 0.0;
            TAR_PTR->bank_control_coeff = 0.0;
            TAR_PTR->pitch_control_coeff = 0.0;
	    TAR_PTR->elevator_trim_coeff = 0.0;
            TAR_PTR->throttle_coeff = 0.0;
            TAR_PTR->after_burner_state = False;
            TAR_PTR->after_burner_power_coeff = 0.0;
            TAR_PTR->engine_power = SRC_PTR->engine_power;
            TAR_PTR->total_mass = SRC_PTR->dry_mass + SRC_PTR->fuel;
	    /* Add up fuel tanks' mass. */
	    for(i = 0; i < SRC_PTR->total_external_fueltanks; i++)
	    {
		sar_external_fueltank_struct *eft_ptr = SRC_PTR->external_fueltank[i];
		if((eft_ptr == NULL) ? 0 : (eft_ptr->flags & SAR_EXTERNAL_FUELTANK_FLAG_ONBOARD))
		    TAR_PTR->total_mass += (eft_ptr->dry_mass + eft_ptr->fuel);
	    }
/* Add up mass of passengers? */
            TAR_PTR->attitude_change_rate.heading =
                SRC_PTR->attitude_change_rate.heading;
            TAR_PTR->attitude_change_rate.pitch =
                SRC_PTR->attitude_change_rate.pitch;
            TAR_PTR->attitude_change_rate.bank =
                SRC_PTR->attitude_change_rate.bank;
            TAR_PTR->attitude_leveling_rate.heading = 0.0;
            TAR_PTR->attitude_leveling_rate.pitch = SRC_PTR->pitch_leveling;
            TAR_PTR->attitude_leveling_rate.bank = SRC_PTR->bank_leveling;
            TAR_PTR->air_brakes_state = ((SRC_PTR->air_brakes_state > 0) ?
                True : False
            );
            TAR_PTR->air_brakes_rate = 0.0;
            /* We'll do all collision crash checks, so set 
             * can_crash_into_other and can_cause_crash to False so
             * FDM does not check for collisions.
             */
            TAR_PTR->can_crash_into_other = False;
            TAR_PTR->can_cause_crash = False;
            if(cb == NULL)
            {
                /* Assume values for unavailable contact bounds values. */
                TAR_PTR->crash_contact_shape =
                    SFMCrashContactShapeSpherical;
                TAR_PTR->crashable_size_radius = 1.0;
                TAR_PTR->crashable_size_z_min = -0.5;
                TAR_PTR->crashable_size_z_max = 0.5;
            }
            else
            {
                switch(cb->contact_shape)
                {
                  case SAR_CONTACT_SHAPE_CYLENDRICAL:
                    TAR_PTR->crash_contact_shape = 
                        SFMCrashContactShapeCylendrical;
                    break;

                  default:      /* Assume SAR_CONTACT_SHAPE_SPHERICAL. */
                    TAR_PTR->crash_contact_shape = 
                        SFMCrashContactShapeSpherical;
                    break;
                }

                TAR_PTR->crashable_size_radius = cb->contact_radius;
                TAR_PTR->crashable_size_z_min = cb->contact_h_min;
                TAR_PTR->crashable_size_z_max = cb->contact_h_max;
            }
            TAR_PTR->touch_down_crash_resistance = 7.0 *
                MAX(option.damage_resistance_coeff, 1.0);
/* Need to work on this. */
            TAR_PTR->collision_crash_resistance = 2.0;

            /* Do not allow attitude leveling if not flyable. */
            if(SRC_PTR->air_worthy_state != SAR_AIR_WORTHY_FLYABLE)
                TAR_PTR->flags &= ~(SFMFlagAttitudeLevelingRate);


#undef TAR_PTR
#undef SRC_PTR
            break;

/* Add SFM values setting from other types of objects here when needed. */

        }

        /* Was the SFM structure obtained from the above checks? */
        if(fdm != NULL)
        {
	    /* Set other common SFM values from the object's general
	     * structure here.
	     */
#define SRC_PTR fdm
#define TAR_PTR obj_ptr



#undef TAR_PTR
#undef SRC_PTR
        }

        return;
}

/*
 *      Gets the flyable object's SFM values to the object.
 */
void SARSimGetSFMValues(
        sar_scene_struct *scene,
        sar_object_struct *obj_ptr
)
{
        SFMModelStruct *fdm = NULL;
        sar_object_aircraft_struct *obj_aircraft_ptr = NULL;
	sar_object_ground_struct *obj_ground_ptr = NULL;


        if((scene == NULL) ||
           (obj_ptr == NULL) 
        )
            return;
            
        switch(obj_ptr->type)
        {
          case SAR_OBJ_TYPE_AIRCRAFT:
            obj_aircraft_ptr = (sar_object_aircraft_struct *)obj_ptr->data;
            if(obj_aircraft_ptr == NULL)
                break;
                
            fdm = obj_aircraft_ptr->fdm;
            if(fdm == NULL)
                break;
                
#define TAR_PTR obj_aircraft_ptr
#define SRC_PTR fdm

            TAR_PTR->center_to_ground_height =
                SRC_PTR->center_to_ground_height;

            TAR_PTR->vel.x = SRC_PTR->velocity_vector.x;
            TAR_PTR->vel.y = SRC_PTR->velocity_vector.y;
            TAR_PTR->vel.z = SRC_PTR->velocity_vector.z;

            TAR_PTR->speed = SRC_PTR->speed;
            TAR_PTR->landed = SRC_PTR->landed_state;

#undef TAR_PTR
#undef SRC_PTR
            break;

	  case SAR_OBJ_TYPE_GROUND:
	    obj_ground_ptr = (sar_object_ground_struct *)obj_ptr->data;
	    if(obj_ground_ptr == NULL)
		break;

#define TAR_PTR obj_ground_ptr
#define SRC_PTR fdm

/* Add code to update the ground object's heightfield heading
 * sin and cos values if/when ground objects have fdm's.
 */
	    fdm = NULL;

#undef TAR_PTR
#undef SRC_PTR
	    break;
        }

        /* Was the SFM structure obtained from the above checks? */
        if(fdm != NULL)
        {   
#define SRC_PTR fdm
#define TAR_PTR obj_ptr
            /* Set other common SFM values to the object's general
             * structure here.
             */
	    sar_contact_bounds_struct *cb;

            TAR_PTR->pos.x = SRC_PTR->position.x;
            TAR_PTR->pos.y = SRC_PTR->position.y;
            TAR_PTR->pos.z = SRC_PTR->position.z;

            TAR_PTR->dir.heading = SRC_PTR->direction.heading;
            TAR_PTR->dir.pitch = SRC_PTR->direction.pitch;
            TAR_PTR->dir.bank = SRC_PTR->direction.bank;

	    /* If object is not flyable and on water (implying it has
	     * crashed into the water) we need to lower the position
	     * so that the object appears submersed.
	     */
	    if((obj_aircraft_ptr == NULL) ?
		0 : obj_aircraft_ptr->on_water
	    )
	    {
		if(obj_aircraft_ptr->air_worthy_state == SAR_AIR_WORTHY_NOT_FLYABLE)
		{
		    TAR_PTR->pos.z = TAR_PTR->ground_elevation_msl;
		}
	    }

	    /* Update contact bounds rotation if rectangular. */
	    cb = TAR_PTR->contact_bounds;
	    if((cb == NULL) ?
		0 : (cb->contact_shape == SAR_CONTACT_SHAPE_RECTANGULAR)
	    )
	    {
		cb->cos_heading = cos(-TAR_PTR->dir.heading);
		cb->sin_heading = sin(-TAR_PTR->dir.heading);
	    }
#undef TAR_PTR
#undef SRC_PTR
        }

        return;
}


/*
 *      Applies natural forces to object.
 *
 *	Note that natural forces applied to flyable objects will be
 *	applied by the SFM and not here.
 */
int SARSimApplyNaturalForce(
	sar_core_struct *core_ptr, sar_object_struct *obj_ptr
)
{
	int tar_obj_num;
        sar_scene_struct *scene;
	sar_direction_struct *dir;
        sar_position_struct *pos, *vel;
	sar_object_helipad_struct *obj_helipad_ptr;
        sar_object_explosion_struct *obj_explosion_ptr;
	sar_object_human_struct *obj_human_ptr;
	sar_object_fueltank_struct *obj_fueltank_ptr;


        if((core_ptr == NULL) || (obj_ptr == NULL))
            return(0);

        scene = core_ptr->scene;
        if(scene == NULL)
            return(0);

        /* Handle by object type. */
        switch(obj_ptr->type)
        {
	  case SAR_OBJ_TYPE_HELIPAD:
	    obj_helipad_ptr = (sar_object_helipad_struct *)obj_ptr->data;
	    if(obj_helipad_ptr == NULL)
		break;

	    /* Helipad follows a referance object? */
	    if(obj_helipad_ptr->flags & SAR_HELIPAD_FLAG_FOLLOW_REF_OBJECT)
	    {
		sar_object_struct *ref_obj_ptr = SARObjGetPtr(
		    core_ptr->object, core_ptr->total_objects,
		    obj_helipad_ptr->ref_object
		);
		if(ref_obj_ptr != NULL)
		{
		    /* Check referance object type, we only move to the
		     * referance object if it is an object type that moves
		     * around (since the helipad is already moved to the
		     * correct referance position when it is loaded).
		     */
		    if((ref_obj_ptr->type == SAR_OBJ_TYPE_AUTOMOBILE) ||
                       (ref_obj_ptr->type == SAR_OBJ_TYPE_WATERCRAFT) ||
                       (ref_obj_ptr->type == SAR_OBJ_TYPE_AIRCRAFT) ||
                       (ref_obj_ptr->type == SAR_OBJ_TYPE_HUMAN)
		    )
		    {
			SARSimWarpObjectRelative(
			    scene, obj_ptr,
			    core_ptr->object, core_ptr->total_objects,
			    obj_helipad_ptr->ref_object,
			    &obj_helipad_ptr->ref_offset,
			    &obj_helipad_ptr->ref_dir
			);
		    }
		}
	    }
	    break;

          case SAR_OBJ_TYPE_SMOKE:
            pos = &obj_ptr->pos;

            /* Smoke puffs float up, so adjust vertical position. */
            pos->z += (7.0 * time_compensation * time_compression);

	    /* Realize changes in position. */
	    SARSimWarpObject(
		scene, obj_ptr,
		pos, NULL
	    );
            break;

          case SAR_OBJ_TYPE_EXPLOSION:
            obj_explosion_ptr = (sar_object_explosion_struct *)obj_ptr->data;
            if(obj_explosion_ptr == NULL)
                break;

            pos = &obj_ptr->pos;

            /* Time to increment explosion animation frame? */
            if(obj_explosion_ptr->next_frame_inc <= cur_millitime)
            {
		int tex_num = obj_explosion_ptr->tex_num;

		/* Set next time to increment animation frame. */
                obj_explosion_ptr->next_frame_inc = cur_millitime +
                    obj_explosion_ptr->frame_inc_int;

		/* Increment current animation frame. */
                obj_explosion_ptr->cur_frame++;

		/* Is explosion object's texture structure available? */
                if(SARIsTextureAllocated(scene, tex_num))
                {
                    v3d_texture_ref_struct *t = scene->texture_ref[tex_num];

		    /* Animation frames cycled? */
                    if(obj_explosion_ptr->cur_frame >= t->total_frames)
		    {
                        obj_explosion_ptr->cur_frame = 0;
			obj_explosion_ptr->frame_repeats++;

			/* Total animation frame cycle repeats exceeded? */
			if((obj_explosion_ptr->frame_repeats >=
			    obj_explosion_ptr->total_frame_repeats) &&
                           (obj_explosion_ptr->total_frame_repeats > 0)
			)
			{
			    /* Mark this object to be deleted by setting
			     * its life span to 1 ms which will cause it to
			     * be deleted.
			     */
			    obj_ptr->life_span = 1;

			    /* Set current frame to be the last frame. */
			    obj_explosion_ptr->cur_frame = t->total_frames - 1;
			}
		    }
		    /* Sanitize current animation frame. */
                    if(obj_explosion_ptr->cur_frame < 0)
                        obj_explosion_ptr->cur_frame = 0;
                }
            }

            /* Get explosion's reference object, that's the object it
	     * is following.
	     */
	    tar_obj_num = obj_explosion_ptr->ref_object;
            if(SARObjIsAllocated(
                core_ptr->object, core_ptr->total_objects,
		tar_obj_num
            ))
            {
		double ground_elevation_msl;
                sar_object_struct *tar_obj_ptr = core_ptr->object[tar_obj_num];

		/* Move explosion object's position to the position of
		 * the referanced object.
		 */
                memcpy(pos, &tar_obj_ptr->pos, sizeof(sar_position_struct));

		/* Get ref object's ground elevation or its z position
		 * (whichever is greater).
		 */
                ground_elevation_msl = MAX(
                    tar_obj_ptr->ground_elevation_msl,
                    tar_obj_ptr->pos.z
                );

		/* Keep above ground level by gravity offset. */
		switch(obj_explosion_ptr->gravity_offset)
		{
		  case SAR_EXPLOSION_GRAVITY_BASE:
		    /* Need to move the position up. */
		    pos->z += obj_explosion_ptr->radius;
		    /* Sanitize z position to base. */
		    if(pos->z < (ground_elevation_msl +
			obj_explosion_ptr->radius)
		    )
			pos->z = ground_elevation_msl +
			    obj_explosion_ptr->radius;
		    break;

		  default:	/* Assume SAR_EXPLOSION_GRAVITY_CENTER */
		    /* Sanitize z position to center. */
		    if(pos->z < ground_elevation_msl)
                        pos->z = ground_elevation_msl;
                    break;
		}
                /* Realize changes in position. */
                SARSimWarpObject(
                    scene, obj_ptr,
                    pos, NULL
                );
            }
            break;

          case SAR_OBJ_TYPE_HUMAN:
	    obj_human_ptr = (sar_object_human_struct *)obj_ptr->data;
	    if(obj_human_ptr == NULL)
		break;

            pos = &obj_ptr->pos;

	    /* Move human only if it is not being moved by something else
	     * we check the human's SAR_HUMAN_FLAG_GRIPPED flag. If it
	     * is set then we do not apply any forces to the human.
	     */
	    if(obj_human_ptr->flags & SAR_HUMAN_FLAG_GRIPPED)
	    {
		/* Human is being moved by another object, possibly the
		 * it is in a rescue basket. In any case we do not apply
		 * any forces to the human object.
		 */
	    }
	    else
	    {
		/* Pull human down with gravity (use constant 3 meters per
		 * second velocity for now).
		 */
		pos->z -= (3.0 * time_compensation * time_compression);

		/* Keep human above ground level. */
		if(pos->z < obj_ptr->ground_elevation_msl)
		    pos->z = obj_ptr->ground_elevation_msl;

		/* Realize changes in position. */
		SARSimWarpObject(
		    scene, obj_ptr,
		    pos, NULL
		);
	    }
	    break;

          case SAR_OBJ_TYPE_FUELTANK:
            obj_fueltank_ptr = (sar_object_fueltank_struct *)obj_ptr->data;
            if(obj_fueltank_ptr == NULL)
                break;

	    dir = &obj_ptr->dir;
            pos = &obj_ptr->pos;

	    /* Landed on ground yet? */
	    if(obj_fueltank_ptr->flags & SAR_FUELTANK_FLAG_ON_GROUND)
	    {

	    }
	    else
	    {
		double	cos_heading = cos(dir->heading),
			sin_heading = sin(dir->heading);

		vel = &obj_fueltank_ptr->vel;

		vel->z += -(
		    SAR_GRAVITY * time_compensation * time_compression
		);
		if(vel->z < obj_fueltank_ptr->vel_z_max)
		    vel->z = obj_fueltank_ptr->vel_z_max;

		pos->x += (((sin_heading * vel->y) +
		    (cos_heading * vel->x)) *
		    time_compensation * time_compression
		);
		pos->y += (((cos_heading * vel->y) -
                    (sin_heading * vel->x)) *
		    time_compensation * time_compression
                );
                pos->z += vel->z * time_compensation * time_compression;

                /* Keep fuel tank above ground level. */
                if(pos->z < obj_ptr->ground_elevation_msl)
		{
                    pos->z = obj_ptr->ground_elevation_msl;

		    /* Mark as landed. */
		    obj_fueltank_ptr->flags |= SAR_FUELTANK_FLAG_ON_GROUND;

                    /* Set life-span when landed on ground so it gets
                     * removed after a while.
                     */
                    obj_ptr->life_span = cur_millitime +
                        option.fuel_tank_life_span;
		}
                /* Realize changes in position. */
                SARSimWarpObject(
                    scene, obj_ptr,
                    pos, NULL
                );
	    }
	    break;

/* Add code to handle other types of objects that need natural forces
 * applied to it.
 */
        }
            
        return(0);
}

/*
 *      Called once per cycle for each object to apply artificial
 *      forces to the object.
 *
 *      Can return 1 to indicate object has been deleted.
 */
int SARSimApplyArtificialForce(
        sar_core_struct *core_ptr, sar_object_struct *obj_ptr
)
{
        int i, air_worthy_state;
        double thrust_output = 0;  
        double ground_elevation, cen_to_gear_height;
        sar_scene_struct *scene;
        sar_direction_struct *dir;
        sar_position_struct *pos, *vel;
        sar_object_aircraft_struct *obj_aircraft_ptr;
        sar_object_human_struct *obj_human_ptr;
        sar_obj_rotor_struct *rotor_ptr;
        sar_obj_landing_gear_struct *lgear_ptr;
        sar_obj_hoist_struct *hoist_ptr;
        sar_contact_bounds_struct *cb;  
        
  
        if((core_ptr == NULL) ||
           (obj_ptr == NULL)
        )
            return(0);

        scene = core_ptr->scene;
        if(scene == NULL)
            return(0);


        /* Get pointers to position and direction structs from object. */
        pos = &obj_ptr->pos;
        dir = &obj_ptr->dir;

        /* Get pointer to contact bounds structure on object (can
         * be NULL so check it later).
         */
        cb = obj_ptr->contact_bounds;


        /* Begin calculating vertical distance from 0.0 to object for
	 * object types that need to land or walk on surfaces (for
	 * instance do not calculate for SAR_OBJ_TYPE_STATIC or
	 * SAR_OBJ_TYPE_GROUND type objects).
	 *
	 * Set starting elevation to 0.0 (not scene's msl elevation,
	 * that is only added when displaying current MSL altitudes).
	 */
	ground_elevation = 0.0;

	/* Check objects of types that need to know about ground
	 * elevation.
	 */
        if((obj_ptr->type == SAR_OBJ_TYPE_AIRCRAFT) ||
           (obj_ptr->type == SAR_OBJ_TYPE_HUMAN) ||
	   (obj_ptr->type == SAR_OBJ_TYPE_FUELTANK)
        )
        {
            /* We need to check if this object has contact bounds that
	     * specify SAR_CRASH_FLAG_SUPPORT_SURFACE. Because if it
	     * does then checking for surface at the position of this
	     * object may hit itself as a surface and cause recursion.
	     */
            if(!(((cb == NULL) ? 0 : cb->crash_flags) &
                SAR_CRASH_FLAG_SUPPORT_SURFACE)
            )
            {
		/* This object does not specify countact bounds for
		 * SAR_CRASH_FLAG_SUPPORT_SURFACE.
		 */

                /* Check for other objects `under' this object for
		 * increased ground elevation, this checks for all
                 * all objects that have either a contact bounds set for
                 * SAR_CRASH_FLAG_SUPPORT_SURFACE and objects of type
                 * SAR_OBJ_TYPE_GROUND.
                 *
                 * Only SAR_OBJ_TYPE_GROUND objects will be treated as
		 * `solid' ground while objects with contact bounds will
		 * be `hollow'.
                 */
                ground_elevation += SARSimFindGround(
                    scene,
                    core_ptr->object, core_ptr->total_objects,
                    pos			/* Position of our object. */
                );
            }
        }
	/* Set new ground elevation to object. */
	obj_ptr->ground_elevation_msl = ground_elevation;


        /* Handle by object type. */
        switch(obj_ptr->type)
        {
          case SAR_OBJ_TYPE_AIRCRAFT:
            obj_aircraft_ptr = (sar_object_aircraft_struct *)obj_ptr->data;
            if(obj_aircraft_ptr == NULL)
                break;

            vel = &obj_aircraft_ptr->vel;
            air_worthy_state = obj_aircraft_ptr->air_worthy_state;

            /* Get pointer to first landing gear. */
            lgear_ptr = ((obj_aircraft_ptr->total_landing_gears > 0) ?
                obj_aircraft_ptr->landing_gear[0] : NULL
            );

            /* Calculate center of object to lower of landing gear height. */
            cen_to_gear_height = obj_aircraft_ptr->belly_height;
            if(lgear_ptr != NULL)
                cen_to_gear_height += ((lgear_ptr->flags & SAR_LGEAR_FLAG_STATE) ?
                    0 : obj_aircraft_ptr->gear_height
                );

            /* Any fuel left? */
            if(obj_aircraft_ptr->fuel > 0.0)
            {
                /* Handle engine initialization. */
                if(obj_aircraft_ptr->engine_state == SAR_ENGINE_STATE_INIT)
                {
                    /* Time to initialize? */
                    if(obj_aircraft_ptr->next_engine_on <= cur_millitime)
                    {
                        obj_aircraft_ptr->engine_state = SAR_ENGINE_STATE_ON;
                        obj_aircraft_ptr->next_engine_on = 0;
                    }
                }
          
                /* Calculate thrust output and consume fuel. */
                if((obj_aircraft_ptr->service_ceiling > 0.0) &&
                   (obj_aircraft_ptr->engine_state == SAR_ENGINE_STATE_ON)
                )   
                {
                    /* Has fuel left, consume. */
                    obj_aircraft_ptr->fuel -= obj_aircraft_ptr->fuel_rate
                        * obj_aircraft_ptr->throttle
                        * time_compensation * time_compression;
                    if(obj_aircraft_ptr->fuel <= 0.0)   
                    {
                        obj_aircraft_ptr->fuel = 0.0;
                        obj_aircraft_ptr->engine_state = SAR_ENGINE_STATE_OFF;
                    }
                    else
                    {
                        /* Calculate thrust output, the closer the
                         * object is to its service ceiling the less
                         * thrust output will be.
                         */
                        thrust_output = obj_aircraft_ptr->throttle *
                            obj_aircraft_ptr->engine_power *
                            MAX(1.0 - (pos->z /
                                obj_aircraft_ptr->service_ceiling),
                                0.0
                            );
                        if(thrust_output < 0.0)
                            thrust_output = 0.0;
                    }
                }
            }
            else if(obj_aircraft_ptr->engine_state == SAR_ENGINE_STATE_ON)
            {
                /* No fuel and engine state still on, in which case turn
                 * it off.
                 */
                obj_aircraft_ptr->engine_state = SAR_ENGINE_STATE_OFF;
            }


            /* Rotate main rotor based on throttle and engine. */
            if((obj_aircraft_ptr->engine_state == SAR_ENGINE_STATE_ON) ||
	       (obj_aircraft_ptr->throttle > 0.0)
	    )
            {
                for(i = 0; i < obj_aircraft_ptr->total_rotors; i++)
                {
                    rotor_ptr = obj_aircraft_ptr->rotor[i];
                    if(rotor_ptr == NULL)
                        continue;

                    rotor_ptr->anim_pos += (int)MAX(
                        (double)(obj_aircraft_ptr->throttle + 0.2) *
                        (double)((int)((sar_grad_anim_t)-1) * 1.40) *
                        time_compensation * time_compression,
                        1
                    );
                    rotor_ptr->rotor_wash_anim_pos += (int)MAX(
                        (double)(obj_aircraft_ptr->throttle + 0.2) *
                        (double)((int)((sar_grad_anim_t)-1) * 0.50) *
                        time_compensation * time_compression,
                        1
                    );
                }
            }

            /* Animate all landing gears. */
            for(i = 0; i < obj_aircraft_ptr->total_landing_gears; i++)
            {
                lgear_ptr = obj_aircraft_ptr->landing_gear[i];
                if(lgear_ptr == NULL)
                    continue;
             
                SARSimUpdateLandingGear(lgear_ptr);
            }     

            /* Animate rescue door. */
            SARSimUpdateDoor(&obj_aircraft_ptr->rescue_door);


            /* Move hoist along with its object. */
            hoist_ptr = SARObjGetHoistPtr(obj_ptr, 0, NULL);
            if((hoist_ptr == NULL) ? 0 : (hoist_ptr->rope_cur > 0.0))
            {
		double a[3 * 1], r[3 * 1];	/* Hoist offset matrices. */
		sar_position_struct *basket_pos = &hoist_ptr->pos;
		int tar_obj_num;
                sar_object_struct *tar_obj_ptr;


		/* Get matrix a as hoist offset translation. */
		a[0] = hoist_ptr->offset.x;
		a[1] = hoist_ptr->offset.y;
		a[2] = hoist_ptr->offset.z;

		/* Rotate matrix a into r. */
		MatrixRotateBank3(a, dir->bank, r);
		MatrixRotatePitch3(r, dir->pitch, a);
		MatrixRotateHeading3(a, dir->heading, r);

                /* Calculate position of hoist's rescue basket in
		 * world coordinates by adding rorated offset relative to
		 * object's center.
		 */
		basket_pos->x = pos->x + r[0];
		basket_pos->y = pos->y + r[1];
		basket_pos->z = pos->z + r[2];

		/* Move basket position down by rope length. */
		basket_pos->z -= hoist_ptr->rope_cur;

                /* Is rope out so much that basket is embedded into
		 * the ground?
		 */
                if(basket_pos->z < ground_elevation)
		{
		    /* Retract rope so that basket is not embedded into
		     * the ground.
		     */
		    hoist_ptr->rope_cur = MAX(
			hoist_ptr->rope_cur -
			ground_elevation + basket_pos->z,
			0.0
		    );

		    /* Sanitize basket z position. */
                    basket_pos->z = ground_elevation;
		}

                /* Adjust heading of hoist basket to match that of
                 * the object it is connected to.
                 */
                memcpy(
                    &hoist_ptr->dir,
                    dir,
                    sizeof(sar_direction_struct)
                );


                /* Move occupants in rescue basket to the new position
		 * of the rescue basket.
		 */
                for(i = 0; i < hoist_ptr->total_occupants; i++)
                {
                    tar_obj_num = hoist_ptr->occupant[i];
                    if(SARObjIsAllocated(
                        core_ptr->object, core_ptr->total_objects,
                        tar_obj_num
                    ))   
                        tar_obj_ptr = core_ptr->object[tar_obj_num];
                    else
                        continue;

		    /* Move passenger along with rescue basket. */
                    memcpy(
                        &tar_obj_ptr->pos,
                        basket_pos,
                        sizeof(sar_position_struct)
                    );
                    /* Move passenger up a bit. */
                    tar_obj_ptr->pos.z += 0.2;

                    /* Direction of occupant is to match direction of
                     * hoist rescue basket.
                     */
                    memcpy(
                        &tar_obj_ptr->dir,
                        &hoist_ptr->dir,
                        sizeof(sar_direction_struct)
                    );

		    /* Formally realize changes to occupant object. */
		    SARSimWarpObject(
			scene, tar_obj_ptr,
			&tar_obj_ptr->pos, &tar_obj_ptr->dir
		    );
                }
            }
            break;

          case SAR_OBJ_TYPE_HUMAN:
            obj_human_ptr = (sar_object_human_struct *)obj_ptr->data;
            if(obj_human_ptr == NULL)
                break;

            /* Update animation. */
            obj_human_ptr->anim_pos += (sar_grad_anim_t)(
                (double)obj_human_ptr->anim_rate *
                time_compensation * time_compression   
            );

/* Handle by flags on human. */

            break;
        }

        return(0);
}

/*
 *      Apply game controls (gctl) values to flyable object.
 */
void SARSimApplyGCTL(
        sar_core_struct *core_ptr,
	sar_object_struct *obj_ptr
)
{
        const double camera_turn_rate = (0.30 * PI *
            CYCLE_LAPSE_US / 1000000.0);

	int rotor_num;
        int air_worthy_state, engine_state;
        sar_scene_struct *scene;
        sar_direction_struct *dir;
        sar_position_struct *pos;
        sar_object_aircraft_struct *obj_aircraft_ptr = NULL;
        sar_contact_bounds_struct *cb;
        double contact_radius = 0.0;
        sar_obj_hoist_struct *hoist_ptr = NULL;
        sar_obj_door_struct *rescue_door_ptr = NULL;
	sar_obj_rotor_struct *rotor_ptr;
        gctl_struct *gc;
        SFMModelStruct *fdm = NULL;
        double  h_con_coeff = 0.0, p_con_coeff = 0.0, b_con_coeff = 0.0,
                throttle_coeff = 0.0;


        if((core_ptr == NULL) ||
           (obj_ptr == NULL)
        )
            return;

        gc = core_ptr->gctl;
        if(gc == NULL)
            return;

        scene = core_ptr->scene;
        if(scene == NULL)
            return;

	/* Get contact bounds structure of the object. */
        cb = obj_ptr->contact_bounds;

	/* Get flat contact radius in meters. */
	contact_radius = SARSimGetFlatContactRadius(obj_ptr);


	/* Apply game controller values based on object type. */            
        switch(obj_ptr->type)
        {
          case SAR_OBJ_TYPE_AIRCRAFT:
            obj_aircraft_ptr = (sar_object_aircraft_struct *)obj_ptr->data;
            if(obj_aircraft_ptr == NULL)
                break;

            dir = &obj_ptr->dir;
            pos = &obj_ptr->pos;
            air_worthy_state = obj_aircraft_ptr->air_worthy_state;
            engine_state = obj_aircraft_ptr->engine_state;

            rescue_door_ptr = &obj_aircraft_ptr->rescue_door;
            hoist_ptr = SARObjGetHoistPtr(obj_ptr, 0, NULL);
            fdm = obj_aircraft_ptr->fdm;


            /* Check if landed on ground. */
            if(obj_aircraft_ptr->landed)
            {
                /* Is landed. */

                /* Handle flight control by airworthy states. */
                if(air_worthy_state == SAR_AIR_WORTHY_FLYABLE)
                {
                    if(engine_state == SAR_ENGINE_STATE_ON)
                    {
                        h_con_coeff = CLIP(gc->heading, -1.0, 1.0);
                        p_con_coeff = CLIP(gc->pitch, -1.0, 1.0);
                        b_con_coeff = CLIP(gc->bank, -1.0, 1.0);

                        throttle_coeff = CLIP(gc->throttle, 0.0, 1.0);
                    }
                    else
                    {
                        h_con_coeff = CLIP(gc->heading, -1.0, 1.0);
                    }
                }
                else if(air_worthy_state == SAR_AIR_WORTHY_OUT_OF_CONTROL)
                {
                    /* Out of control and landed. */
 
                    /* Check if engines are on, if so give it just  
                     * half power at most.
                     */
                    if(engine_state == SAR_ENGINE_STATE_ON)
                        throttle_coeff = CLIP(gc->throttle * 0.5, 0.0, 1.0);
                }
                else
                {
                    /* Not flyable and landed. */
                }
            }
            else
            {   
                /* In flight. */
                
                /* Handle flight control by airworthy states. */
                if(air_worthy_state == SAR_AIR_WORTHY_FLYABLE)
                {
                    h_con_coeff = CLIP(gc->heading, -1.0, 1.0);
                    p_con_coeff = CLIP(gc->pitch, -1.0, 1.0);
                    b_con_coeff = CLIP(gc->bank, -1.0, 1.0);
         
                    /* Check if engines are on. */
                    if(engine_state == SAR_ENGINE_STATE_ON)
                        throttle_coeff = CLIP(gc->throttle, 0.0, 1.0);
                }
                else if(air_worthy_state == SAR_AIR_WORTHY_OUT_OF_CONTROL)
                {
                    /* In flight, but out of control. */

		    switch(obj_aircraft_ptr->flight_model_type)
		    {
		      case SAR_FLIGHT_MODEL_AIRPLANE:
                        /* Aircraft, roll when out of control. */
                        h_con_coeff = CLIP(gc->heading, -0.1, 0.1); 
                        p_con_coeff = CLIP(gc->pitch, -0.1, 0.1);
                        b_con_coeff = CLIP(gc->bank + 1.1, -1.0, 1.0);
			break;

		      case SAR_FLIGHT_MODEL_HELICOPTER:
                        /* Helicopter, spin when out of control. */
                        h_con_coeff = CLIP(gc->heading + 1.1, -1.0, 1.0);
                        p_con_coeff = CLIP(gc->pitch, -0.1, 0.1);
                        b_con_coeff = CLIP(gc->bank, -0.1, 0.1);
			break;
                    }
            
                    /* Check if engines are on, if so give it less power
                     * (try 0.75 for now).
                     */
                    if(engine_state == SAR_ENGINE_STATE_ON)
                        throttle_coeff = CLIP(gc->throttle * 0.75, 0.0, 1.0);
                }
                else
                {
                    /* In flight and not flyable. */

                }
            }

            /* Change in throttle (value of throttle fetched above)? */
	    if(obj_aircraft_ptr->throttle != throttle_coeff)
	    {
		/* Calculate maximum allowed change in throttle. */
/* Should get throttle change coeff from object. But for now we use constants
 * dependant on the global option.flight_physics.
 */
		double throttle_dc;

		switch(option.flight_physics)
		{
		  case FLIGHT_PHYSICS_REALISTIC:
		    throttle_dc = 0.035 * time_compensation *
			time_compression;
		    break;
		  case FLIGHT_PHYSICS_MODERATE:
		    throttle_dc = 0.06 * time_compensation *
                        time_compression;
		    break;
		  default:
		    throttle_dc = 0.1 * time_compensation *
                        time_compression;
                    break;
		}

		/* Decrease in throttle? */
		if(obj_aircraft_ptr->throttle > throttle_coeff)
		{
		    obj_aircraft_ptr->throttle = MAX(MAX(
			throttle_coeff,
			obj_aircraft_ptr->throttle - throttle_dc
		    ), 0.0);
		}
		/* Increase in throttle? */
		else if(obj_aircraft_ptr->throttle < throttle_coeff)
		{
		    obj_aircraft_ptr->throttle = MIN(MIN(
			throttle_coeff,
			obj_aircraft_ptr->throttle + throttle_dc
		    ), 1.0);
		}
	    }

            /* Air brakes on/off. */
            if(obj_aircraft_ptr->air_brakes_state > -1)
                obj_aircraft_ptr->air_brakes_state =
                    ((gc->air_brakes_state) ? 1 : 0);

            /* Wheel brakes on/off. */
            if(obj_aircraft_ptr->wheel_brakes_state > -1)
            {
                if(gc->wheel_brakes_state == 2)
                    obj_aircraft_ptr->wheel_brakes_state = 2;
                else if(gc->wheel_brakes_coeff > 0.0)
                    obj_aircraft_ptr->wheel_brakes_state = 1;
                else
                    obj_aircraft_ptr->wheel_brakes_state = 0; 
            }

            /* Update SAR FDM control positions. */
            if(fdm != NULL)
            {
                fdm->heading_control_coeff = h_con_coeff;  
                fdm->pitch_control_coeff = p_con_coeff;
		fdm->elevator_trim_coeff = obj_aircraft_ptr->elevator_trim;
                fdm->bank_control_coeff = b_con_coeff;
                fdm->throttle_coeff = obj_aircraft_ptr->throttle;
                fdm->gear_brakes_state = ((obj_aircraft_ptr->wheel_brakes_state > 0) ?
                    True : False
                );
                fdm->air_brakes_state = ((obj_aircraft_ptr->air_brakes_state > 0) ?
                    True : False
                );
            }

	    /* Update rotor control positions. */
	    for(rotor_num = 0; rotor_num < obj_aircraft_ptr->total_rotors; rotor_num++)
	    {
		rotor_ptr = obj_aircraft_ptr->rotor[rotor_num];
		if(rotor_ptr == NULL)
		    continue;

		/* Report values from gc since *_con_coeff reflects those
		 * dependant on engine state and other physics values.
		 */
		SARSimRotorUpdateControls(
		    rotor_ptr,
		    CLIP(gc->pitch, -1.0, 1.0),
		    CLIP(gc->bank, -1.0, 1.0)
		);
	    }
            break;
        }


        /* Move hoist? Check if hoist pointer was fetched from above. */
        if(hoist_ptr != NULL)
        {
            double rope_rate = hoist_ptr->rope_rate;
	    double hoist_up_coeff = ((gc->ctrl_state) ?
		gc->zoom_in_coeff : gc->hoist_up_coeff
	    );
	    double hoist_down_coeff = ((gc->ctrl_state) ?
                gc->zoom_out_coeff : gc->hoist_down_coeff
            );

            if(1)
            {
                /* Hoist up? */
                if(hoist_up_coeff > 0.0)
                {
                    double prev_rope = hoist_ptr->rope_cur;

                    /* Pull in/up. */
                    hoist_ptr->rope_cur -= (rope_rate * time_compensation *
                        time_compression * hoist_up_coeff);
                    if(hoist_ptr->rope_cur < hoist_ptr->contact_z_max)
                    {
                        hoist_ptr->rope_cur = 0;
                
                        if(prev_rope > hoist_ptr->contact_z_max)
                        {
                            SARSimDoHoistIn(core_ptr, obj_ptr);
                    
                            /* Door exists and is currently opened? */
                            if((rescue_door_ptr == NULL) ? 0 :
				(rescue_door_ptr->flags & SAR_DOOR_FLAG_STATE)
			    )
                            {   
				SARSimOpDoor(
				    scene,
				    &core_ptr->object, &core_ptr->total_objects,
				    obj_ptr,
				    0			/* Close. */
				);
                            }
                        }
                    }
                }
		/* Hoist down? */
                else if(hoist_down_coeff > 0.0)
                {
                    /* Open rescue door as needed. */
                    if((rescue_door_ptr == NULL) ? 0 :
			!(rescue_door_ptr->flags & SAR_DOOR_FLAG_STATE)
		    )
                    {
			/* Rescue door exists but is closed, we need to
			 * open it.
			 */
			SARSimOpDoor(
			    scene,
			    &core_ptr->object, &core_ptr->total_objects,
			    obj_ptr,
			    1			/* Open. */
			);

			/* Here is also a good point to check if we have
			 * enough room for additional passengers.
			 */
			if((obj_aircraft_ptr == NULL) ? 0 :
			    (obj_aircraft_ptr->passengers >=
				obj_aircraft_ptr->passengers_max)
			)
			{
			    /* Is this the player object? */
			    if(scene->player_obj_ptr == obj_ptr)
				SARMessageAdd(
				    scene,
				    SAR_MESG_NO_ROOM_LEFT_FOR_PASSENGERS
				);
			}
                    }

                    /* Hoist was in? */
                    if(hoist_ptr->rope_cur < hoist_ptr->contact_z_max)
                    {
                        double  tx = hoist_ptr->offset.x,
                                ty = hoist_ptr->offset.y,
                                tz = hoist_ptr->offset.z;
                                
                        hoist_ptr->rope_cur = hoist_ptr->contact_z_max;
                                        
                        /* Move to initial position at hoist center. */
                        SFMOrthoRotate2D(
                            obj_ptr->dir.heading, &tx, &ty
                        );
                        hoist_ptr->pos.x = obj_ptr->pos.x + tx;
                        hoist_ptr->pos.y = obj_ptr->pos.y + ty;
                        hoist_ptr->pos.z = obj_ptr->pos.z + tz -
                            hoist_ptr->rope_cur;
                    }
                    else
                    {
                        /* Lower rope. */
                        hoist_ptr->rope_cur += (rope_rate *
                            time_compensation * time_compression *
                            hoist_down_coeff);
                        if(hoist_ptr->rope_cur > hoist_ptr->rope_max)
                            hoist_ptr->rope_cur = hoist_ptr->rope_max;
                    }
                }
            }
        }

        /* ********************************************************* */
        /* Camera look direction. */
        switch(scene->camera_ref)
        {
          case SAR_CAMERA_REF_HOIST:
            dir = &scene->camera_hoist_dir;

            /* Hat moves camera around target in spot referance. */
            if(gc->hat_x != 0)  
            {
                dir->heading = SFMSanitizeRadians(
                    dir->heading - (gc->hat_x * camera_turn_rate *
                    time_compensation)      /* No time acceleration. */
                );
            }
                        
            if(gc->hat_y != 0)
            {
                double prev_pitch = dir->pitch; 
                dir->pitch = SFMSanitizeRadians(
                    dir->pitch - (gc->hat_y * camera_turn_rate *
                    time_compensation)      /* No time acceleration. */
                );
                        
                if((dir->pitch < (1.5 * PI)) &&
                   (dir->pitch > (0.5 * PI))
                )
                {
                    if(prev_pitch < PI)
                        dir->pitch = (0.5 * PI);
                    else
                        dir->pitch = (1.5 * PI);
                }             
            }
        
            /* Zoom in and out. */
            if(!gc->ctrl_state)
            {
                if(gc->zoom_in_coeff > 0)  
                {               
                    scene->camera_hoist_dist -= ((gc->shift_state) ?
                        50 : 10 
                    ) * time_compensation * time_compression *
                        SAR_SEC_TO_CYCLE_COEFF * gc->zoom_in_coeff;
                    
                    if(scene->camera_hoist_dist < contact_radius)
                        scene->camera_hoist_dist = contact_radius;
                }
                if(gc->zoom_out_coeff > 0)
                {
                    scene->camera_hoist_dist += ((gc->shift_state) ?
                        50 : 10
                    ) * time_compensation * time_compression *
                        SAR_SEC_TO_CYCLE_COEFF * gc->zoom_out_coeff;
                }
            }
                        
            break;
                   
          case SAR_CAMERA_REF_MAP:
            pos = &scene->camera_pos;
            /* Hat moves map location. */
            if(gc->hat_x != 0)
            {
                pos->x += (gc->hat_x *
                    ((gc->shift_state) ? 25000 : 5000)
                    * time_compensation
                );
            }
            if(gc->hat_y != 0) 
            {
                pos->y += (gc->hat_y *     
                    ((gc->shift_state) ? 25000 : 5000)
                    * time_compensation
                );
            }
                        
            /* Zoom in and out (adjust z position). */
            if(!gc->ctrl_state)
            {
                if(gc->zoom_in_coeff > 0)
                {
                    pos->z -= ((gc->shift_state) ?
                        10000 : 500
                    ) * time_compensation * time_compression *
                        SAR_SEC_TO_CYCLE_COEFF * gc->zoom_in_coeff;
                        
                    if(pos->z < 10)
                        pos->z = 10;
                }       
                if(gc->zoom_out_coeff > 0)
                {  
                    pos->z += ((gc->shift_state) ?
                        10000 : 500  
                    ) * time_compensation * time_compression *
                        SAR_SEC_TO_CYCLE_COEFF * gc->zoom_out_coeff;
                    if(pos->z > SFMFeetToMeters(SAR_MAX_MAP_VIEW_HEIGHT))
                        pos->z = SFMFeetToMeters(SAR_MAX_MAP_VIEW_HEIGHT);
                }
            }
            break;
             
          case SAR_CAMERA_REF_TOWER:
            pos = &scene->camera_pos;
                
/* Work on this later. */
                    
            break;
             
          case SAR_CAMERA_REF_SPOT:
            dir = &scene->camera_spot_dir;
 
            /* Hat moves camera around target in spot referance. */
            if(gc->hat_x != 0)
            {
                dir->heading = SFMSanitizeRadians(
                    dir->heading - (gc->hat_x * camera_turn_rate *
                    time_compensation)      /* No time acceleration. */
                );
            }           
                    
            if(gc->hat_y != 0)
            {
                double prev_pitch = dir->pitch;
                dir->pitch = SFMSanitizeRadians(
                    dir->pitch - (gc->hat_y * camera_turn_rate *
                    time_compensation)      /* No time acceleration. */
                );
                        
                if((dir->pitch < (1.5 * PI)) &&
                   (dir->pitch > (0.5 * PI))
                )
                {
                    if(prev_pitch < PI)
                       dir->pitch = (0.5 * PI);
                    else
                        dir->pitch = (1.5 * PI);
                }
            }
                    
/*
            dir->bank = 0;
 */
            /* Zoom in and out. */
            if(!gc->ctrl_state)
            {
                if(gc->zoom_in_coeff > 0)
                {
                    scene->camera_spot_dist -= ((gc->shift_state) ?
                        50 : 10
                    ) * time_compensation * time_compression *
                        SAR_SEC_TO_CYCLE_COEFF * gc->zoom_in_coeff;
                        
                    if(scene->camera_spot_dist < contact_radius)
                        scene->camera_spot_dist = contact_radius;
                }
                if(gc->zoom_out_coeff > 0)
                {
                    scene->camera_spot_dist += ((gc->shift_state) ?
                        50 : 10
                    ) * time_compensation * time_compression *
                        SAR_SEC_TO_CYCLE_COEFF * gc->zoom_out_coeff;
                }
            }
            break;

          default:      /* Default to cockpit. */
            dir = &scene->camera_cockpit_dir;   

            if(gc->shift_state)
            {
                /* Look in right angle directions (no change in cam pitch). */

                if(gc->hat_x > 0.5)
                    dir->heading = (0.5 * PI);
                else if(gc->hat_x < -0.5)
                    dir->heading = (1.5 * PI);

                if(gc->hat_y > 0.5)
                    dir->heading = (0.0 * PI);
                else if(gc->hat_y < -0.5)
                    dir->heading = (1.0 * PI);
            }
            else
            {
                /* Adjust cam heading and pitch. */
                    
                if(gc->hat_x != 0)
                {
                    dir->heading = SFMSanitizeRadians(
                        dir->heading + (gc->hat_x * camera_turn_rate *
                        time_compensation)
                    );
                }
                        
                if(gc->hat_y != 0)
                {
                    double prev_pitch = dir->pitch;
                    dir->pitch = SFMSanitizeRadians(
                        dir->pitch - (gc->hat_y * camera_turn_rate *
                        time_compensation)      
                    );
            
                    /* Sanitize. */
                    if((dir->pitch < (1.5 * PI)) &&
                       (dir->pitch > (0.5 * PI))
                    )
                    {
                        if(prev_pitch < PI)
                            dir->pitch = (0.5 * PI);
                        else
                            dir->pitch = (1.5 * PI);
                    }
                }

/*              dir->bank = 0;  Camera bank is unaffected. */
            }
            break;
        }

        return;
}
