#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>

#include "../include/os.h"

#ifdef __MSW__
# include <windows.h>
#endif
#include <GL/gl.h>

#include "../include/tga.h"

#include "v3dhf.h"


void V3DHFNormalByCrossGridVectors(
        double v1i, double v1k,
        double v2j, double v2k
);
double V3DHFGetZPixelCoeff(
        u_int32_t *ptr,
        int x, int y,
        int width, int height
);
int V3DHFLoadFromFile(
        const char *path,       /* Heightfield image file. */
        double x_len, double y_len, double z_len,       /* Size. */
        int *width_rtn, int *height_rtn,        /* In grids or pixels. */
        double *x_spacing_rtn,  /* Spacing between grids (points). */
        double *y_spacing_rtn,     
        u_int8_t **data_rtn,    /* Dynamically allocated z points
                                 * can be NULL.
                                 */
        void *gl_list,          /* GL list (can be NULL). */
        u_int32_t options       /* Any of V3D_HF_OPT_*. */
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))



/*
 *	Sets normal given the vectors along the grid (left hand rule).
 *
 *	v1 is the vector running left to right along the grid's
 *	`x axis'.
 *
 *	v2 is the vector running fore to aft along the grid's
 *	`y axis'.
 *
 *	i and j are the value along the 2D grid, and k is the
 *	`height'.
 */
void V3DHFNormalByCrossGridVectors(
	double v1i, double v1k,
	double v2j, double v2k
)
{
	double r[3];	/* Resultant. */
	double m;	/* Magnitude of resultant. */

	r[0] = (double)(-v1k * v2j);
	r[1] = (double)(-v1i * v2k);
	r[2] = (double)(v1i * v2j);

	m = hypot(hypot(r[0], r[1]), r[2]);
	if(m > 0.0)
	    glNormal3f(
		(GLfloat)(r[0] / m),
		(GLfloat)(r[2] / m),
		(GLfloat)(-r[1] / m)
	    );


	return;
}

/*
 *	Returns the z pixel value as a coefficient from 0.0 to 1.0,
 *	taking into account only the first 8 bits of the 32 bit pixel.
 *
 *	The given values are assumed valid.
 */
double V3DHFGetZPixelCoeff(
	u_int32_t *ptr,
	int x, int y,
	int width, int height
)
{
	return((double)((u_int32_t)(ptr[(y * width) + x]) & 0x000000ff) /
	       (double)0xff
	);
}

/*
 *	Loads a heightfield from an greyscale tga image file specified by
 *	path.
 *
 *	If data_rtn is not NULL then the pointer will be set to
 *	a newly allocated array of z points (each z point is 8 bits)
 *	and the specified width and height will be updated (in units of
 *	grids, which would also be the units of pixels).
 *
 *	If gl_list is not NULL (not (GLint)0) then OpenGL operations
 *	to draw the heightfield will be recorded. gl_list should have been
 *	created by the calling function and the calling function is
 *	responsable for ending it.
 *
 *	Center of heightfield will be centered at x_len / 2 and
 *	y_len / 2, and the base of the heightfield will be at z = 0.
 *	and highest point (when a pixel value is 0xff) will be z_len.
 *
 *	Returns non zero on error.
 */
int V3DHFLoadFromFile(
        const char *path,       /* Heightfield image file. */
        double x_len, double y_len, double z_len,       /* Size. */
	int *width_rtn, int *height_rtn,	/* In grids or pixels. */
        double *x_spacing_rtn,	/* Spacing between grids. */
        double *y_spacing_rtn,
        u_int8_t **data_rtn,    /* Dynamically allocated z points
                                 * can be NULL.
                                 */
        void *gl_list,          /* GL list (can be NULL). */
        u_int32_t options       /* Any of V3D_HF_OPT_*. */
)
{
	int status, total_z_points;
	double x_len_half, y_len_half;
	u_int8_t *data_ptr = NULL;	/* Local data pointer. */
	u_int32_t *img_data;
	struct stat stat_buf;
	tga_data_struct td;


	/* Reset return sizes if possable. */
	if(width_rtn != NULL)
	    (*width_rtn) = 0;
        if(height_rtn != NULL)
            (*height_rtn) = 0;
	if(data_rtn != NULL)
	    (*data_rtn) = NULL;
	if(x_spacing_rtn != NULL)
	    (*x_spacing_rtn) = 0.0;
	if(y_spacing_rtn != NULL)
	    (*y_spacing_rtn) = 0.0;


	/* Check inputs. */
	if(path == NULL)
	    return(-1);

	if((x_len <= 0) ||
           (y_len <= 0) ||
	   (z_len <= 0)
	)
	    return(-3);

	/* Calculate half sizes. */
	x_len_half = x_len / 2;
	y_len_half = y_len / 2;

	/* File exists? */
        if(stat(path, &stat_buf))
        {
            fprintf(stderr,
                "%s: No such file.\n",
                path
            );
            return(-1);
        }
#ifdef S_ISDIR
	if(S_ISDIR(stat_buf.st_mode))
	{
            fprintf(stderr,
                "%s: Is a directory.\n",
                path
            );
            return(-1);
	}
#endif	/* S_ISDIR */

        /* Load data from file, read as 32 bits. */
        status = TgaReadFromFile(
            path,
            &td,
            32		/* Read to 32 bits. */
        );
        if(status != TgaSuccess)
        {
            TgaDestroyData(&td);
            return(-1);
	}

        /* Check if size is big enough. */
        if(td.width < 2)
            fprintf(stderr,
 "%s: Warning: Image size is too small in width.\n",
                path
            );
        if(td.height < 2)
            fprintf(stderr,
 "%s: Warning: Image size is too small in height.\n",
                path
            );

	/* Record size. */
        if(width_rtn != NULL)  
            (*width_rtn) = MAX((int)td.width, 0);
        if(height_rtn != NULL)
            (*height_rtn) = MAX((int)td.height, 0);

	/* Calculate total z points. */
	total_z_points = MAX((int)td.width, 0) * MAX((int)td.height, 0);

        /* Set image pointer to loaded image data. */
        img_data = (u_int32_t *)td.data;
        if(img_data == NULL)
        {   
            TgaDestroyData(&td);
            return(-1);  
        }

	/* Allocate z points array if input data_rtn is not NULL
	 * and there are z points is non-empty.
	 */
	if((data_rtn != NULL) && (total_z_points > 0))
	{
	    data_ptr = (u_int8_t *)malloc(
		total_z_points * sizeof(u_int8_t)
	    );
	    (*data_rtn) = data_ptr;

	    if(data_ptr == NULL)
		fprintf(stderr,
 "V3DHFLoadFromFile(): Cannot allocate memory to load `%s' which has %i heightfield points.\n",
                    path, total_z_points
                );
	}


	/* Begin setting z points and/or recording GLlist. */
	if(total_z_points > 0)
	{
	    int cx, cy;
	    double x_sp, y_sp;	/* Each grid (pixel) is this many meters. */
	    int img_w, img_h;	/* Number of grids. */
	    GLfloat vz[4];	/* Four corner z values. */

	    /* Get image size. */
	    img_w = (int)td.width;
	    img_h = (int)td.height;

	    /* Calculate spacing. */
	    if(img_w > 0.0)
		x_sp = x_len / (double)img_w;
	    else
		x_sp = 0.0;

	    if(img_h > 0.0)
		y_sp = y_len / (double)img_h;
	    else
		y_sp = 0.0;

            /* Set grid spacing returns if possable. */
	    if(x_spacing_rtn != NULL)
		(*x_spacing_rtn) = x_sp;
            if(y_spacing_rtn != NULL)
                (*y_spacing_rtn) = y_sp;

	    /* Set z point data return if allocated. */
	    if(data_ptr != NULL)
	    {
		/* X and y number of grids should match pixels on image. */
		u_int8_t *data_tmp_ptr = data_ptr;
		u_int32_t	*img_ptr = img_data,
				*img_ptr_end = img_data + (td.width * td.height);

		while(img_ptr < img_ptr_end)
		    *data_tmp_ptr++ = (u_int8_t)((*img_ptr++) & 0x000000ff);
	    }

            glBegin(GL_TRIANGLES);

	    /* Go through each `top edge' grid. */
	    for(cy = 0; cy < img_h; cy++)
	    {
		/* Go through each `left edge' grid. */
                for(cx = 0; cx < img_w; cx++)
                {
		    /* Get z pizel height of each `corner'.
		     * Vertex numbering layout:
		     *
		     *   v1---v2
		     *   | \   |
		     *   |  \  |
                     *   |   \ |
                     *   v3---v4
		     *
		     * Each grid consists of two triangles, an upper
		     * right one and a lower left one.
		     */

		    /* Upper left (vertex #1 always valid). */
		    vz[0] = (GLfloat)(V3DHFGetZPixelCoeff(
			img_data, cx, cy, img_w, img_h
		    ) * z_len);

		    /* Upper right (vertex #2). */
		    if((cx + 1) < img_w)
		    {
                        vz[1] = (GLfloat)(V3DHFGetZPixelCoeff(
                            img_data, cx + 1, cy, img_w, img_h
                        ) * z_len);
		    }
		    else
		    {
			/* Data for upper right not available, use upper left. */
			vz[1] = vz[0];
		    }

		    /* Lower left (vertex #3). */
                    if((cy + 1) < img_h)
		    {
                        vz[2] = (GLfloat)(V3DHFGetZPixelCoeff(
                            img_data, cx, cy + 1, img_w, img_h
                        ) * z_len);
		    }
                    else
		    {
			/* Data for lower left not available, use upper left. */
                        vz[2] = vz[0];
		    }

		    /* Lower right (vertex #4). */
                    if(((cx + 1) < img_w) &&
                       ((cy + 1) < img_h)
		    )
		    {
                        vz[3] = (GLfloat)(V3DHFGetZPixelCoeff(
                            img_data, cx + 1, cy + 1, img_w, img_h
                        ) * z_len);
		    }
		    /* Lower left valid? */
		    else if((cy + 1) < img_h)
		    {
			vz[3] = vz[2];
		    }
		    /* Upper right valid? */
                    else if((cx + 1) < img_w)
		    {
                        vz[3] = vz[1];
		    }
		    /* Upper left and upper right not available, use upper left. */
		    else
		    {
			vz[3] = vz[0];
		    }

		    /* Begin recording GL drawing, check winding. */
		    if(options & V3D_HF_OPT_COUNTER_CLOCKWISE)
		    {
			/* Winding counter clockwise. */

			/* Upper right triangle. */
                        /* Set normal. */
                        V3DHFNormalByCrossGridVectors(
                            x_sp, vz[1] - vz[0],
                            y_sp, vz[1] - vz[3]
                        );

                        /* Vertex #1. */
                        glTexCoord2f(
                            (GLfloat)((cx + 0) * x_sp / x_len),
                            (GLfloat)((cy + 0) * y_sp / y_len)
                        );
                        glVertex3f(
                            (GLfloat)(((cx + 0) * x_sp) - x_len_half),
                            (GLfloat)(vz[0]),   
                            (GLfloat)(((cy + 0) * y_sp) - y_len_half)
                        );

                        /* Vertex #4. */
                        glTexCoord2f(
                            (GLfloat)((cx + 1) * x_sp / x_len),
                            (GLfloat)((cy + 1) * y_sp / y_len)
                        );
                        glVertex3f(
                            (GLfloat)(((cx + 1) * x_sp) - x_len_half),  
                            (GLfloat)(vz[3]),
                            (GLfloat)(((cy + 1) * y_sp) - y_len_half)
                        );

                        /* Vertex #2. */
                        glTexCoord2f(
                            (GLfloat)((cx + 1) * x_sp / x_len),
                            (GLfloat)((cy + 0) * y_sp / y_len)
                        );
                        glVertex3f(
                            (GLfloat)(((cx + 1) * x_sp) - x_len_half),
                            (GLfloat)(vz[1]),
                            (GLfloat)(((cy + 0) * y_sp) - y_len_half)
                        );


                        /* Lower left triangle. */
                        /* Set normal. */
                        V3DHFNormalByCrossGridVectors(
                            x_sp, vz[3] - vz[2],
                            y_sp, vz[0] - vz[2]
                        );

                        /* Vertex #1. */
                        glTexCoord2f(
                            (GLfloat)((cx + 0) * x_sp / x_len),
                            (GLfloat)((cy + 0) * y_sp / y_len)
                        );
                        glVertex3f(
                            (GLfloat)(((cx + 0) * x_sp) - x_len_half),
                            (GLfloat)(vz[0]),
                            (GLfloat)(((cy + 0) * y_sp) - y_len_half)
                        );

                        /* Vertex #3. */
                        glTexCoord2f(
                            (GLfloat)((cx + 0) * x_sp / x_len),
                            (GLfloat)((cy + 1) * y_sp / y_len)
                        );
                        glVertex3f(
                            (GLfloat)(((cx + 0) * x_sp) - x_len_half),
                            (GLfloat)(vz[2]),   
                            (GLfloat)(((cy + 1) * y_sp) - y_len_half)
                        );

                        /* Vertex #4. */
                        glTexCoord2f(
                            (GLfloat)((cx + 1) * x_sp / x_len),
                            (GLfloat)((cy + 1) * y_sp / y_len)
                        );
                        glVertex3f(
                            (GLfloat)(((cx + 1) * x_sp) - x_len_half),
                            (GLfloat)(vz[3]),
                            (GLfloat)(((cy + 1) * y_sp) - y_len_half)
                        );
		    }
		    else
		    {
			/* Winding clockwise. */

                        /* Upper right triangle. */
                        /* Set normal. */
                        V3DHFNormalByCrossGridVectors(
                            x_sp, vz[1] - vz[0],
                            y_sp, vz[1] - vz[3]
                        );

                        /* Vertex #1. */
                        glTexCoord2f(
                            (GLfloat)((cx + 0) * x_sp / x_len),
                            (GLfloat)((cy + 0) * y_sp / y_len)
                        );
                        glVertex3f(  
                            (GLfloat)(((cx + 0) * x_sp) - x_len_half),
                            (GLfloat)(vz[0]),
                            (GLfloat)(((cy + 0) * y_sp) - y_len_half)
                        );

                        /* Vertex #2. */
                        glTexCoord2f(   
                            (GLfloat)((cx + 1) * x_sp / x_len),
                            (GLfloat)((cy + 0) * y_sp / y_len) 
                        );
                        glVertex3f(
                            (GLfloat)(((cx + 1) * x_sp) - x_len_half),
                            (GLfloat)(vz[1]),
                            (GLfloat)(((cy + 0) * y_sp) - y_len_half)
                        );

                        /* Vertex #4. */
                        glTexCoord2f(
                            (GLfloat)((cx + 1) * x_sp / x_len),
                            (GLfloat)((cy + 1) * y_sp / y_len)
                        );
                        glVertex3f(
                            (GLfloat)(((cx + 1) * x_sp) - x_len_half),
                            (GLfloat)(vz[3]),
                            (GLfloat)(((cy + 1) * y_sp) - y_len_half)
                        );


                        /* Lower left triangle. */
                        /* Set normal. */
                        V3DHFNormalByCrossGridVectors(
                            x_sp, vz[3] - vz[2],
                            y_sp, vz[0] - vz[2]
                        );

                        /* Vertex #1. */
                        glTexCoord2f(
                            (GLfloat)((cx + 0) * x_sp / x_len),
                            (GLfloat)((cy + 0) * y_sp / y_len)
                        );
                        glVertex3f(
                            (GLfloat)(((cx + 0) * x_sp) - x_len_half),
                            (GLfloat)(vz[0]),   
                            (GLfloat)(((cy + 0) * y_sp) - y_len_half)
                        );

                        /* Vertex #4. */
                        glTexCoord2f(
                            (GLfloat)((cx + 1) * x_sp / x_len),
                            (GLfloat)((cy + 1) * y_sp / y_len)
                        );
                        glVertex3f(
                            (GLfloat)(((cx + 1) * x_sp) - x_len_half),
                            (GLfloat)(vz[3]),
                            (GLfloat)(((cy + 1) * y_sp) - y_len_half)
                        );

                        /* Vertex #3. */
                        glTexCoord2f(
                            (GLfloat)((cx + 0) * x_sp / x_len),
                            (GLfloat)((cy + 1) * y_sp / y_len)
                        );
                        glVertex3f(
                            (GLfloat)(((cx + 0) * x_sp) - x_len_half),
                            (GLfloat)(vz[2]),
                            (GLfloat)(((cy + 1) * y_sp) - y_len_half)
                        );

		    }	/* Begin recording GL drawing, check winding. */


		}
	    }

            glEnd();
	}


	/* Deallocate loaded image data. */
        TgaDestroyData(&td);

	return(0);
}
